from rest_framework import serializers
from arkindex.documents.models import Corpus
from arkindex.dataimport.models import DataImport, Repository, Revision
import gitlab.v4.objects


class RevisionSerializer(serializers.ModelSerializer):
    """
    Serialize a repository revision
    """

    date = serializers.DateTimeField(source='created')
    commit_url = serializers.URLField(read_only=True)

    class Meta:
        model = Revision
        fields = (
            'id',
            'date',
            'hash',
            'message',
            'author',
            'commit_url',
            'repo_id',
        )


class RepositorySerializer(serializers.ModelSerializer):
    """
    Serialize a repository
    """
    enabled = serializers.BooleanField(read_only=True)

    class Meta:
        model = Repository
        fields = (
            'id',
            'url',
            'enabled',
            'corpus',
        )
        extra_kwargs = {
            'id': {'read_only': True},
            'url': {'read_only': True},
        }


class RepositoryStartImportSerializer(serializers.ModelSerializer):
    """
    A serializer used by the RepositoryStartImport endpoint to return a DataImport ID.
    This serializer is required to get the OpenAPI schema generation to work.
    """

    import_id = serializers.UUIDField(source='id')

    class Meta:
        model = DataImport
        fields = ('import_id',)
        read_only_fields = ('import_id',)


class ExternalRepositorySerializer(serializers.Serializer):
    """
    Serialize a Git repository from an external API
    """

    # Useless fields, added to prevent 500 errors when opening
    # an API endpoint in the browser
    id = serializers.IntegerField(min_value=0)
    corpus = serializers.UUIDField()

    def to_representation(self, obj):
        if isinstance(obj, gitlab.v4.objects.Project):
            return {
                "id": obj.id,
                "name": obj.name_with_namespace,
                "url": obj.web_url,
            }
        else:
            raise NotImplementedError

    def to_internal_value(self, data):
        """
        Deserializing only requires a 'id' attribute
        """
        if not data.get('id'):
            raise serializers.ValidationError({
                'id': 'This field is required.'
            })
        if not data.get('corpus'):
            raise serializers.ValidationError({
                'corpus': 'This field is required.'
            })

        return {
            'id': data['id'],
            'corpus': Corpus.objects.writable(self.context['request'].user)
                                    .get(id=data['corpus'])
        }
